"""
run_file(workflowFilePath::String, system::OrchestratorRegistry.ServiceRegistry)

Executes a workflow defined in a file, leveraging a provided `ServiceRegistry` to manage the services involved.

This function automates the setup, execution, and teardown of a containerized workflow, simplifying the process of running complex workflows with multiple dependencies. It performs the following steps:
- Parses the workflow file specified by `workflowFilePath` to extract tasks.
- Retrieves the `FACTResultsIO` service from the provided `ServiceRegistry` to configure the results container.
- Manages workflow execution, including handling tasks, updating global variables, and interacting with services.
- Optionally updates the results container with the final workflow results.

# Arguments
- `workflowFilePath::String`: The path to the file containing the workflow definition.
- `system::OrchestratorRegistry.ServiceRegistry`: A registry of services required for workflow execution.

# Behavior
- Parses the workflow file to determine the tasks to execute.
- Uses the `FACTResultsIO` service from the registry to set up a results container.
- Executes tasks defined in the workflow, interacting with the necessary services as defined in the registry.
- Manages cleanup and updates for the results container after workflow execution.

# Side Effects
- Starts and stops Docker containers as required for workflow execution.
- Modifies the results container based on workflow outcomes.

# Example
To run a workflow defined in a file located at `/path/to/workflow.yml`, using a preconfigured service registry `system`:

> run_file("/path/to/workflow.yml", system)

This will execute the workflow, managing all service interactions and container operations as needed.
"""
function run_file(workflowFilePath::String, system::OrchestratorRegistry.ServiceRegistry)
    # Parse the workflow file
    tasks = parseWorkflow(workflowFilePath)

    # Create a reference to results container
    services = OrchestratorRegistry.get_services_by_name(system, "FACT_unified_DB")
    if isempty(services)
        error("Service 'FACT_unified_DB' not found in the registry")
    end
    service = services[1]
    results_container = Dict(
        "user" => "root",
        "password" => "devops",
        "host" => service.host,
        "port" => service.port
    )

    # Execute the workflow
    manage_workflow(tasks, results_container, system)
end

"""
manage_workflow(tasks_list::Vector{Dict}, results_container::Dict, system::OrchestratorRegistry.ServiceRegistry)

Executes a workflow by managing the execution of tasks based on their dependencies and the provided service registry.

This function handles a workflow where tasks are represented as a list of dictionaries. Each task is executed in parallel using multithreading, respecting dependencies between tasks. Tasks communicate their completion status via channels, and their execution may update the results container.

# Arguments
- tasks_list::Vector{Dict}: A list of task dictionaries. Each dictionary represents a task and includes the following keys:
  - id (String): The unique identifier for the task.
  - dependencies (Vector{String}): A list of task IDs that must complete before this task can run.
  - Other relevant task-specific information (e.g., service, function, arguments).
- results_container::Dict: A dictionary representing the results container, used by tasks to store or retrieve results.
- system::OrchestratorRegistry.ServiceRegistry: The service registry used to manage services required by tasks during execution.

# Returns
- None

# Behavior
1. Converts the tasks_list into a dictionary (Dict{String, Dict}) where the task id is the key for efficient access.
2. Initializes channels for each task to signal dependency completion.
3. Spawns a thread for each task:
   - Waits for dependencies to complete before execution.
   - Executes the task, updating the results_container and other relevant data structures.
4. Ensures all tasks are completed before the function exits.

# Example
```julia
tasks_list = [
    Dict(
        "id" => "task1",
        "dependencies" => [],
        "service" => "example_service",
        "function" => "run_task1",
        "arguments" => "arg1",
        "parameters_dict" => Dict()
    ),
    Dict(
        "id" => "task2",
        "dependencies" => ["task1"],
        "service" => "example_service",
        "function" => "run_task2",
        "arguments" => "arg2",
        "parameters_dict" => Dict()
    )
]

results_container = Dict(
    "user" => "root",
    "password" => "password",
    "host" => "localhost",
    "port" => 5432
)

system = ServiceRegistry()  # Assuming ServiceRegistry is preconfigured

manage_workflow(tasks_list, results_container, system)
```
"""
function manage_workflow(tasks_list::Vector{<:Dict}, results_container::Dict, system::OrchestratorRegistry.ServiceRegistry)
    # Convert tasks_list into a dictionary keyed by the "id" field for easy dependency resolution
    # Each element of tasks_list is assumed to have a unique "id" key.
    tasks = Dict(task["id"] => task for task in tasks_list)

    overwrite = true
    update_condition = Atomic{Bool}(false)  # Flag for signaling updates
    
    # Initialize Channels for each task to signal completion
    task_channels = Dict{String, Channel{Bool}}()
    for task_id in keys(tasks)
        task_channels[task_id] = Channel{Bool}(1)
    end

    task_handles = Task[]  # Array of task handles for spawned threads

    for (task_id, task_info) in tasks
        t = @spawn begin
            try
                # Prepare dependencies information if present
                dependencies_info = Dict{String, Dict}()
                if haskey(task_info, "dependencies") && !isempty(task_info["dependencies"])
                    for dep in task_info["dependencies"]
                        if haskey(tasks, dep)
                            dependencies_info[dep] = tasks[dep]
                        else
                            # Dependency not found warning (optional)
                            # println("Warning: Dependency $dep not found in tasks.")
                        end
                    end
                end

                # Wait for dependencies to complete if there are any
                if haskey(task_info, "dependencies") && !isempty(task_info["dependencies"])
                    for dep in task_info["dependencies"]
                        # Wait until the dependent task signals completion
                        take!(task_channels[dep])
                        # Put the signal back for other tasks that might depend on the same dependency
                        put!(task_channels[dep], true)
                    end
                end

                # Execute the task
                run_task(task_id, task_info, dependencies_info, system, results_container, update_condition, overwrite)

            catch ex
                # Handle exceptions by printing error messages and stacktrace
                println("Error occurred in task $task_id: ", ex)
                println("Stacktrace: ", stacktrace(catch_backtrace()))

            finally
                # Signal that this task is complete
                put!(task_channels[task_id], true)
            end
        end
        push!(task_handles, t)
    end

    # Wait for all tasks to complete
    foreach(wait, task_handles)
    println("All tasks completed.")
end