"""
parseWorkflow(filePath::String)

Parses a workflow configuration file in TOML format into a flat list of dictionaries. Each task is represented as a dictionary, including an `id` key for its name.

# Arguments
- filePath::String: The path to the TOML file containing the workflow configuration.

# Returns
- Vector{Dict}: A list of task dictionaries. Each dictionary represents a task and includes:
  - id (String): The unique identifier for the task (extracted from the task name).
  - Other key-value pairs corresponding to task-specific details from the TOML file.

# Behavior
1. Reads and parses the specified TOML file.
2. Ensures the file contains a "tasks" section.
3. Converts each task entry into a dictionary, adding an `id` field corresponding to the task name.
4. Returns a list of these dictionaries.

# Example
```julia
# Example TOML file structure:
[tasks.no_counties]
service = "FACT_geo"
function = "get_geo_objects"
arguments = "no_counties"
parameters_dict = {}
dependencies = []

[tasks.another_task]
service = "Another_service"
function = "another_function"
arguments = "arg_value"
parameters_dict = {key1 = "value1"}
dependencies = ["no_counties"]

# Usage:
tasks = parseWorkflow("path/to/workflow.toml")
for task in tasks
    println("Task ID: ", task["id"])
    println("Service: ", task["service"])
end
``
"""
function parseWorkflow(filePath::String)
    # Parse the TOML file
    data = TOML.parsefile(filePath)

    # Ensure the "tasks" section exists
    if !haskey(data, "tasks")
        error("Invalid workflow file: 'tasks' section not found.")
    end

    # Initialize tasks as a vector of Dict
    tasks = Dict{String, Any}[]

    for (task_name, task_details) in data["tasks"]
        if task_details isa Dict
            # Add the task name as the "id" and flatten into a single dictionary
            task_details["id"] = task_name
            push!(tasks, task_details)
        end
    end

    # Validate that at least one task is parsed
    if isempty(tasks)
        error("No valid tasks found in the workflow file.")
    end

    return tasks
end
