# Define a global cache dictionary. The key is a tuple of parameters, and the value is the DataFrame result.
# This cache is used to avoid redundant database queries for task dependencies.
const RESULTS_CACHE = Dict{Tuple{String, String}, DataFrame}()

"""
run_task(
    task_name::String,
    task_info::Dict,
    dependencies_info::Dict,
    system::OrchestratorRegistry.ServiceRegistry,
    results_container::Dict,
    update_condition::Atomic{Bool},
    overwrite::Bool
)

Executes a single task within a workflow, processing its dependencies and updating the results container as needed.

This function performs the following steps:
1. Processes task parameters, resolving dependencies from the provided `dependencies_info`.
2. Executes the core work of the task using the `task_work` function.
3. Updates the `update_condition` flag if the task modifies the results container.
4. Logs task execution details, including start and end times and the task duration.

# Arguments
- `task_name::String`: The name of the task to execute.
- `task_info::Dict`: A dictionary containing details about the task, including:
  - `"function"`: The name of the function to call for task execution.
  - `"arguments"`: A list of arguments to pass to the task function.
  - `"parameters_dict"`: A dictionary of parameters that may depend on task dependencies.
- `dependencies_info::Dict`: A dictionary containing resolved information about the task's dependencies.
- `system::OrchestratorRegistry.ServiceRegistry`: A service registry used to access services required by the task.
- `results_container::Dict`: A dictionary representing the results container, where the task can store or retrieve results.
- `update_condition::Atomic{Bool}`: An atomic flag to indicate if the results container was updated.
- `overwrite::Bool`: A flag indicating whether existing results should be overwritten.

# Behavior
- Resolves dependencies using `process_task_parameters`.
- Executes the specified task function with the provided arguments and parameters.
- Uses an atomic compare-and-swap (`atomic_cas!`) operation to update the `update_condition` flag if the task modifies the results container.
- Logs start time, end time, and execution duration for the task.

# Side Effects
- May modify the `results_container` if the task produces output.
- Updates the `update_condition` flag if the task modifies the results container.

# Example
Define a task and its dependencies, and run it:

```julia
task_name = "example-task"
task_info = Dict(
    "function" => "example_function",
    "arguments" => ["arg1", "arg2"],
    "parameters_dict" => Dict("param1" => "value1")
)

dependencies_info = Dict(
    "dependency1" => Dict("result" => "dependency-result")
)

results_container = Dict(
    "user" => "root",
    "password" => "devops",
    "host" => "localhost",
    "port" => 8080
)

system = ServiceRegistry()
add_service!(system, Service("123", "example-service", "http://localhost", 9000))

update_condition = Atomic{Bool}(false)
overwrite = true

run_task(task_name, task_info, dependencies_info, system, results_container, update_condition, overwrite)
```

This example demonstrates how to define and execute a task within a workflow, updating the results container as necessary.
"""
function run_task(task_name::String, task_info::Dict, dependencies_info::Dict, system::OrchestratorRegistry.ServiceRegistry, results_container::Dict, update_condition::Atomic{Bool}, overwrite::Bool)
    start_time = now()  # Capture the start time
    println("Starting task $task_name at $(Dates.format(start_time, "HH:MM:SS"))")

    condition_met = false

    # in case of need process parameters dictionary to get results from dependencies in database
    parameters_dictionary = process_task_parameters(task_info["parameters_dict"], dependencies_info, results_container)

    try
        condition_met = task_work(task_info["function"], task_info["arguments"], parameters_dictionary, system, results_container, overwrite)
    catch e
        println("An error occurred in task $task_name: $e")
    end

    # Update condition
    if !update_condition[] && condition_met
        atomic_cas!(update_condition, false, true) # CAS: Compare-And-Swap
    end
    
    end_time = now()  # Capture the end time
    println("Completed task $task_name at $(Dates.format(end_time, "HH:MM:SS"))")

    # Calculate and print the duration
    duration = print_duration_in_hms(end_time - start_time)  # Duration as a String interpretation of a Period type
    println("Duration of task $task_name: $duration")

end

function task_work(functionName::String, arguments::String, parametersDict::Dict, system::OrchestratorRegistry.ServiceRegistry, results_container::Dict, overwrite::Bool)
    condition_met = false
    # Connect to the results container
    results_connection = get_results_connection(results_container)

    does_table_exist = does_table_exist_for_method(results_connection, functionName, arguments, parametersDict)
    
    #println("Does table exist: $does_table_exist for $functionName, $arguments, $parametersDict")

    # Close the connection to the results container
    close_connection(results_connection)

    if !does_table_exist || overwrite
        # Prepare arguments string representation for eval
        args_str = parametersDict |> Dict |> string
        #call_str = "$functionName(\"$arguments\", $args_str)"
        call_str = "$functionName(\"$arguments\", $args_str, system)"
        #println("Executing $call_str")
        # Convert call string to expression and evaluate it
        expr = Meta.parse(call_str)
        #df = eval(expr)
        df = eval(:(let system = $system; $expr end))

        # Connect to the results container
        results_connection = get_results_connection(results_container)
        
        if overwrite
            overwrite_table(results_connection, functionName, arguments, parametersDict, df)
        else 
            write_table(results_connection, functionName, arguments, parametersDict, df)
        end
        
        # Close the connection to the results container
        close_connection(results_connection)
        
        condition_met = true
    end

    return condition_met
end

"""
    process_task_parameters(original_dict::Dict, dependencies_info::Dict, results_container::Dict) -> Dict

Processes a dictionary of task parameters to resolve any references to data produced by other tasks. References are identified
by a specific format (__REF__taskName[!, :columnName]) and are replaced with the actual data values from the task outputs.
This function supports dynamic parameter passing between tasks in a workflow, where the output of one task can be used as an
input parameter for another.

# Arguments
- `original_dict::Dict`: The original dictionary containing the task parameters, potentially including references to other tasks' outputs.
- `dependencies_info::Dict`: A dictionary containing information about the tasks and their dependencies.
- `results_container::Dict`: The information to activate a database connection used to fetch data referenced by task outputs.

# Returns
`Dict`: A new dictionary with all references resolved to actual data values. If a reference cannot be resolved due to missing data or incorrect reference format, an error is raised.

# Example
Given a task parameter dictionary with references to outputs of other tasks, this function will resolve those references:

> process_task_parameters(original_parameters, dependencies_info, results_connection)

This will return a new dictionary where all __REF__ references have been replaced with the actual values from the respective tasks' outputs.
"""
function process_task_parameters(original_dict::Dict, dependencies_info::Dict, results_container::Dict)
    # Define the regular expression pattern to match the reference format
    ref_pattern = r"__REF__(\w+)\[!, :(\w+)\]"

    # Initialize a new dictionary to hold the processed parameters
    new_dict = Dict()

    for (key, value) in original_dict
        if occursin("__REF__", string(value))
            # Extract task name and column name from the reference
            match_result = match(ref_pattern, value)

            if match_result !== nothing
                task_name = String(match_result.captures[1])
                column_name = Symbol(match_result.captures[2])

                # Retrieve the DataFrame using the task name
                df = get_task_df(task_name, dependencies_info, results_container)

                # Check if the DataFrame and the column exist
                if df !== nothing && String(column_name) in names(df)
                    # Update the new dictionary with the column value
                    new_dict[key] = df[!, column_name]
                else
                    error("DataFrame or column not found for reference: $value")
                end
            else
                error("Invalid reference format in value: $value")
            end
        else
            # For non-reference values, just copy them to the new dictionary
            new_dict[key] = value
        end
    end

    return new_dict
end

"""
    get_task_df(task_name::String, dependencies_info::Dict, results_connection::MySQL.Connection) -> DataFrame

Retrieves a DataFrame for a given task by executing a database query through a results connection. 
This function looks up the task's details in `dependencies_info` using `task_name` as the key. If the task is found,
it extracts the function name, arguments, and parameters dictionary from the task's definition and then queries the database 
to fetch the corresponding data.

# Arguments
- `task_name::String`: The name of the task for which data is to be fetched.
- `dependencies_info::Dict`: A dictionary containing task definitions, including function names, arguments, and parameters.
- `results_connection::MySQL.Connection`: An active database connection through which the query will be executed.

# Returns
`DataFrame`: A DataFrame containing the results of the query. Returns an empty DataFrame if the task is not found in `dependencies_info`.

# Example
Assuming `dependencies_info` is a dictionary with task definitions and `results_connection` is an established MySQL connection:

> get_task_df("myTask", dependencies_info, results_connection)

This will return a DataFrame with the data for `myTask`, based on the details specified in `dependencies_info`.
"""
function get_task_df(task_name::String, dependencies_info::Dict, results_container::Dict)

    # Check for a cached result first
    cache_key = (task_name, string(hash(dependencies_info)))  # Generate a unique key for the current parameters
    if haskey(RESULTS_CACHE, cache_key)
        #println("Using cached results for task $task_name")
        return RESULTS_CACHE[cache_key]  # Return the cached DataFrame
    end

    # Connect to the results container
    results_connection = get_results_connection(results_container)

    df = DataFrame()

    if haskey(dependencies_info, task_name)
        task_definition = dependencies_info[task_name]

        # Extract the required information
        functionName = String(get(task_definition, "function", ""))  # Default to empty string if not found
        arguments = String(get(task_definition, "arguments", ""))  # Adjust based on actual key if different
        parametersDict = Dict(get(task_definition, "parameters_dict", Dict()))

        # Assuming get_table is a function that queries the database and returns a DataFrame
        df = get_table(results_connection, functionName, arguments, parametersDict)

        # Cache the result before returning
        RESULTS_CACHE[cache_key] = df
    end

    # Close the connection to the results container
    close_connection(results_connection)

    return df
end

function print_duration_in_hms(duration::Period)
    total_milliseconds = Dates.value(duration)  # Get total milliseconds from the duration
    total_seconds = div(total_milliseconds, 1000)  # Convert milliseconds to seconds
    hours = total_seconds ÷ 3600  # Integer division to get hours
    minutes = (total_seconds % 3600) ÷ 60  # Remaining minutes
    seconds = total_seconds % 60  # Remaining seconds

    # Format the string with leading zeros for minutes and seconds
    return "$(hours)h:$(lpad(minutes, 2, "0"))m:$(lpad(seconds, 2, "0"))s"
end